function [r, dr, d2r] = generate_paramecium(k, n)
    if 2*floor(k/2)+1 > n
        error("n needs to be greater/equal to 2 * floor(k/2) + 1");
    end
    s = linspace(0, 1, n+1);
    s = s(1:end-1);
    y = (1 + 0.1*sin(2*pi*s) - 0.2*cos(2*pi*s) + 0.3*sin(4*pi*s) - 0.2*cos(4*pi*s)).';

    c = splineCoeffPeriodic(y, k);
    r = @(s) splineEval(s*n, c, k, 0);
    dr = @(s) n*splineEval(s*n, c, k, 1);
    d2r = @(s) n^2*splineEval(s*n, c, k, 2);
end


%%%
% spline stuffses
function ys = bSpline(xs, m)
    if (m==0)
        ys = ones(size(xs));
        ys(xs > 0.5) = 0;
        ys(xs <= -0.5) = 0;
    else
        ys = (xs+(m+1)/2)./m.*bSpline(xs+0.5, m-1) ...
             + ((m+1)/2-xs)./m.*bSpline(xs-0.5, m-1);
    end
end

function ys = dlbSpline(xs, m, l)
    if m < l
        ys = zeros(size(xs));
    elseif l == 0
        ys = bSpline(xs, m);
    else
        ys = l/m * (dlbSpline(xs+0.5, m-1, l-1) - dlbSpline(xs-0.5, m-1, l-1)) ...
             + (xs+(m+1)/2)./m.*dlbSpline(xs+0.5, m-1, l) ...
             + ((m+1)/2-xs)./m.*dlbSpline(xs-0.5, m-1, l);
    end
end

function c = splineCoeffPeriodic(y, m)
    n = size(y, 1);
    k = floor(m/2);

    A = toeplitz([bSpline(0:-1:-k, m), zeros(1, n-1-2*k), bSpline(k:-1:1, m)]);

    c = A\y;
    c = [c(end-k+1:end); c; c(1); c(2:k+1)];
end

function ys = splineEval(xs, c, m, l)
    n = size(c, 1);
    k = floor(m/2);

    ys = zeros(size(xs));    
    for ii=1:n
        ys = ys + c(ii) .* dlbSpline(xs-ii+1+k, m, l);
    end
end