function [vd, er] = visdata_double_layer(sbp, n, rhos, g, hd, gap, hs)
    [xb, yb, blimits] = approximate_boundary(sbp, hs);
    [xd, yd, ud] = quasi_inverse_marching_squares(sbp, blimits, hd);
    ud(~isnan(ud)) = 1;
    ud(isnan(ud)) = 0;
    ud = imerode(ud, strel('disk', gap, 0));
    ud(ud == 0) = NaN;

    ub = g([xb; yb]);

    xs = [xd(~isnan(ud)).'; yd(~isnan(ud)).'];
    ue = g(xs);

    u = potential_double_layer(sbp, n, rhos, xs);
    ud(~isnan(ud)) = u.';

    u = u - ue;
    er = norm(u, 'inf');

    vd.xb = xb; vd.yb = yb; vd.ub = ub;
    vd.xd = xd; vd.yd = yd; vd.ud = ud;
end


%%%
% sbp to meshgrid implementation
function [xb, yb, blimits] = approximate_boundary(sbp, h)
    ss = linspace(0, 1, ceil(1/h));
    gammas = sbp.gamma(ss);
    blimits = [min(gammas, [], 2), max(gammas, [], 2)];
    xb = gammas(1, :);
    yb = gammas(2, :);
end

function [xd, yd, ud] = quasi_inverse_marching_squares(sbp, blimits, h)
    blimits = blimits+[-h, h];
    ni = ceil(abs(diff(blimits, 1, 2))/h);
    x0 = (sum(blimits, 2) - h*(ni+1))/2;
    [xd, yd] = ndgrid(linspace(x0(1)+h, x0(1)+h*ni(1), ni(1)), ...
                      linspace(x0(2)+h, x0(2)+h*ni(2), ni(2)));
    ud = ones(size(xd));
    
    s = 0;
    d = h/norm(sbp.dgamma(s));
    in = round((sbp.gamma(s)-x0) / h);
    if all(in > 0) & all(in <= ni)
        ud(in(1), in(2)) = NaN;
    end
    e = 1;
    ip = round((sbp.gamma(s)+e*h*sbp.ngamma(s)-x0) / h);
    while all(ip == in)
        e = 3/2*e;
        ip = round((sbp.gamma(s)+e*h*sbp.ngamma(s)-x0) / h);
    end
    if all(ip > 0) & all(ip <= ni)
        if ud(ip(1), ip(2)) ~=0
            ud(ip(1), ip(2)) = 2;
        end
    end
    io = in;
    while s < 1
        t = mod(s+d, 1);
        in = round((sbp.gamma(t)-x0) / h);
        if all(in == io)
            d = 3/2*d;
        elseif any(abs(io - in) > 1)
            d = d/2;
        else
            s = s+d;
            io = in;
            if all(in > 0) & all(in <= ni)
                ud(in(1), in(2)) = 0;
            end
            e = 1;
            ip = round((sbp.gamma(t)+e*h*sbp.ngamma(t)-x0) / h);
            while all(ip == in)
                e = 3/2*e;
                ip = round((sbp.gamma(t)+e*h*sbp.ngamma(t)-x0) / h);
            end
            if all(ip > 0) & all(ip <= ni)
                if ud(ip(1), ip(2)) ~=0
                    ud(ip(1), ip(2)) = -1;
                end
            end
        end
    end
    l = bwlabel(abs(ud), 4);
    ud(ud == 0) = NaN;
    for kk=unique(l(ud == -1)).'
        ud(l == kk) = NaN;
    end
end